--[[
JIT compiler interface for C++ script dynamic compilation.
]]
local Class = createClass{name="JITCompiler", bases="base.Object"}
local utils = import "base.utils"
local bm = import "base.BindingsManager"
local file = import "pl.file"
local path = import "pl.path"

function Class:__init()
  Class.super.__init(self)
  bm:loadBindings("LLVM", "luaLLVM")
  self.script_search_paths = {}

  local app = nv.NervApp.instance()
  local root_path = app:getRootPath()
  self.script_dir = root_path.."assets/cpp/"
  self.pch_dir = root_path.."compiled/"
  self.bc_dir = root_path.."compiled/"

  self.bin_dir = root_path.."bin/msvc64/"

  self._dynLibs = {}
  self.lambda_count = 0
  self.lambda_map = {}

  self.current_cl_args = {}
  self.current_header_paths = {}
  self.current_macro_defs = {}
  self.current_pch_file = ""

  self.cl_args_hash = ""
  self.header_paths_hash = ""
  self.macro_defs_hash = ""
  self.pch_file_hash = ""

  self.current_context_hash = ""

  self.cl_args_lists = {}
  self.header_paths_lists = {}
  self.macro_defs_lists = {}


  self:addScriptSearchPath(self.script_dir)
  self:init()
end

function Class:addScriptSearchPath(p)
  logDEBUG("JIT: Adding script search path: ", p)
  table.insert(self.script_search_paths, p)
end

-- Method used to register new lists:
function Class:registerCommandLineArgList(clName, args)
  self.cl_args_lists[clName] = args
end

function Class:registerHeaderPathList(name, list)
  self.header_paths_lists[name] = list
end

function Class:registerMacroDefList(name, list)
  self.macro_defs_lists[name] = list
end

-- Get the hash for a given buffer in the current compilation context:
function Class:getContextualBufferHash(buf)
  local hash = nv.sha256_from_buffer(buf)
  return nv.sha256_from_buffer(self.current_context_hash .. hash)
end

-- Methods used to generate a hash value from a string list:
function Class:computeStringListHash(list)
  return nv.sha256_from_buffer(table.concat(list, " "))
end

function Class:updateContextHash()
  self.current_context_hash = self.cl_args_hash .. self.header_paths_hash .. self.macro_defs_hash .. self.pch_file_hash
  -- logDEBUG("Current context hash is: ", self.current_context_hash)
end

function Class:updatePCHHash()
  self.pch_file_hash = nv.sha256_from_buffer(self.current_pch_file)
  self:updateContextHash()
end

function Class:updateMacroDefsHash()
  self.macro_defs_hash = self:computeStringListHash(self.current_macro_defs)
  self:updateContextHash()
end

function Class:updateHeaderPathsHash()
  self.header_paths_hash = self:computeStringListHash(self.current_header_paths)
  self:updateContextHash()
end

function Class:updateCommandLineArgsHash()
  self.cl_args_hash = self:computeStringListHash(self.current_cl_args)
  self:updateContextHash()
end

-- Setup usage of given lists:
function Class:useCommandLineArgLists(listNames, keep)
  
  if not keep then
    self.current_cl_args = {}
  end

  for _,lname in ipairs(listNames) do 
    local plist = self.cl_args_lists[lname]
    for _,entry in ipairs(plist) do 
      table.insert(self.current_cl_args, entry)
    end
  end

  self.jit:setupCommandLine(self.current_cl_args)

  -- This will anyway clear the other settings:
  self.current_header_paths = {}
  self.current_macro_defs = {}

  self:updateCommandLineArgsHash()
  self:updateHeaderPathsHash()
  self:updateMacroDefsHash()
end

function Class:useHeaderPathLists(listNames, keep)
  
  if not keep then
    -- clear the previous list of header search paths:
    self.jit:clearHeaderSearchPaths();
    self.current_header_paths = {}
  end

  for _,lname in ipairs(listNames) do 
    local plist = self.header_paths_lists[lname]
    for _,entry in ipairs(plist) do 
      table.insert(self.current_header_paths, entry[1])
      self.jit:addHeaderSearchPath(entry[1], entry[2]);
    end
  end

  -- We now generate the hash:
  self:updateHeaderPathsHash()
end

function Class:useMacroDefLists(listNames, keep)
  if not keep then
    -- clear the previous list of header search paths:
    self.jit:clearMacroDefinitions();
    self.current_macro_defs = {}
  end

  for _,lname in ipairs(listNames) do 
    local plist = self.macro_defs_lists[lname]
    for _,dval in ipairs(plist) do 
      table.insert(self.current_macro_defs, dval)
      self.jit:addMacroDefinition(dval)
    end
  end

  -- We now generate the hash:
  self:updateMacroDefsHash()
end

function Class:usePCHFile(pchfile)
  self.current_pch_file = pchfile
  self.jit:usePCHFile(self.current_pch_file)
  self:updatePCHHash()
end

function Class:addCommandLineArgs(args)
  for _,argname in ipairs(args) do
    table.insert(self.current_cl_args, argname)
  end

  self.jit:setupCommandLine(self.current_cl_args)

  -- This will anyway clear the other settings:
  self.current_header_paths = {}
  self.current_macro_defs = {}

  self:updateCommandLineArgsHash()
  self:updateHeaderPathsHash()
  self:updateMacroDefsHash()
end

function Class:addHeaderPaths(paths, ptype)  
  ptype = ptype or nv.LLVMHeaderType.ANGLED

  for _,path in ipairs(paths) do
    table.insert(self.current_header_paths, path)
    self.jit:addHeaderSearchPath(path, ptype);
  end

  -- We now generate the hash:
  self:updateHeaderPathsHash()
end

function Class:addMacroDefs(dvals)
  for _,dval in ipairs(dvals) do
    table.insert(self.current_macro_defs, dval)
    self.jit:addMacroDefinition(dval)
  end

  -- We now generate the hash:
  self:updateMacroDefsHash()
end

function Class:init()
  nv.initLLVM()

  self.jit = nv.NervJIT()

  self:registerCommandLineArgList("default", {
    "-triple=x86_64-pc-windows-msvc19.16.27030",
    "-x", "c++",
    "-mrelax-all",
    "-mincremental-linker-compatible",
    "-disable-free",
    -- "-disable-llvm-verifier",
    "-discard-value-names",
    "-mrelocation-model", "pic",
    "-pic-level", "2",
    "-mthread-model", "posix",
    "-mframe-pointer=none",
    "-relaxed-aliasing",
    "-fmath-errno",
    "-fno-rounding-math",
    -- "-masm-verbose", -- for LLVM 10 only
    "-mconstructor-aliases",
    "-munwind-tables",
    "-target-cpu", "x86-64",
    "-mllvm",
    "-x86-asm-syntax=intel",
    "-stack-protector","2",
    "-fcxx-exceptions",
    "-fexceptions",
    "-fexternc-nounwind",
    -- "-flto-visibility-public-std",
    "-fms-volatile",
    "-fdiagnostics-format", "msvc",
    "-dwarf-column-info",
    "-Wall",
    "-std=c++17",
    "-fdeprecated-macro",
    "-ferror-limit","19",
    -- "-fmessage-length", "174", --LLVM 10 only
    "-fmessage-length=174", -- LLVM 11 only
    -- "-fno-use-cxa-atexit",
    -- "-fuse-cxa-atexit",
    -- "-fregister-global-dtors-with-atexit",
    "-fms-extensions",
    "-fms-compatibility",
    "-fms-compatibility-version=19.16.27030",
    "-fdelayed-template-parsing",
    "-fcolor-diagnostics",
    "-fobjc-runtime=gcc",
    -- "-fdiagnostics-show-option", LLVM 10 only
    "-faddrsig",
    -- "-cc1",
    -- "-cc1",
    -- "--dependent-lib=msvcrt",
    -- "-femulated-tls",
    -- "-fno-emulated-tls",
  })

  self:registerHeaderPathList("default", {
    {"W:/Projects/NervSeed/deps/msvc64/llvm-20200409/lib/clang/11.0.0/include", nv.LLVMHeaderType.SYSTEM},
    {"D:/Apps/VisualStudio2017_CE/VC/Tools/MSVC/14.16.27023/include", nv.LLVMHeaderType.SYSTEM},
    {"D:/Apps/VisualStudio2017_CE/VC/Tools/MSVC/14.16.27023/atlmfc/include", nv.LLVMHeaderType.SYSTEM},
    {"C:/Program Files (x86)/Windows Kits/10/Include/10.0.18362.0/ucrt", nv.LLVMHeaderType.SYSTEM},
    {"C:/Program Files (x86)/Windows Kits/10/Include/10.0.18362.0/shared", nv.LLVMHeaderType.SYSTEM},
    {"C:/Program Files (x86)/Windows Kits/10/Include/10.0.18362.0/um", nv.LLVMHeaderType.SYSTEM},
    {"C:/Program Files (x86)/Windows Kits/10/Include/10.0.18362.0/winrt", nv.LLVMHeaderType.SYSTEM},
  })

  self:registerMacroDefList("default", {
    "_MT",
    "_DLL",
    "_CRT_SECURE_NO_WARNINGS",
    "_SILENCE_CXX17_ITERATOR_BASE_CLASS_DEPRECATION_WARNING",
    "__EXCEPTIONS",
    "_CPPUNWIND",
  })

  logDEBUG("JIT: Setting up JIT from command line...")
  self:useCommandLineArgLists{"default"}

  logDEBUG("JIT: Setting up header paths...")
  self:useHeaderPathLists{"default"}
  
  logDEBUG("JIT: Setting up macro definitions...")
  self:useMacroDefLists{"default"}

  -- Ensure the PCH hash is updated too:
  self:updatePCHHash()

  -- Next we link to the crt libraries:
  logDEBUG("JIT: Linking CRT libraries...")
  local crt_path = self.bin_dir.."Microsoft.VC141.CRT/"
	self:addDynamicLib(crt_path.."msvcp140.dll");
	self:addDynamicLib(crt_path.."msvcp140_2.dll");
	self:addDynamicLib(crt_path.."concrt140.dll");
	self:addDynamicLib(crt_path.."vcruntime140.dll"); 
  self:addDynamicLib(self.bin_dir.."dlls/ucrtbase.dll");
  -- self:addDynamicLib(self.bin_dir.."dlls/kernel32.dll");
  self:addDynamicLib(self.bin_dir.."llvm_syms.dll");
  
  -- **WARNING**: should not try to link with the MSVCRT dll file:
  -- this will lead to memory issues when trying to ise ostringstream for instance. 
  -- self:addDynamicLib(crt_path.."MSVCRT.dll");

  self:addDynamicLib("C:/WINDOWS/system32/kernel32.dll");
  
  -- check if we can build some C++ script here:
  -- self.jit:loadModuleFromBuffer([[
    --   #include <iostream>
    
    --   extern "C" void showMessage() {
      --       std::cout << "Hello world from my C++ script!" << std::endl;
      --   }
      -- ]])
      -- self.jit:call("showMessage")
      
  -- We add the missing extensions we need into the lua bindings here:
  self:addDynamicLib(self.bin_dir.."nvCore.dll");
  self:addDynamicLib(self.bin_dir.."lua51.dll");
  self:addHeaderPaths{
    "W:/Projects/NervSeed/deps/msvc64/LuaJIT-2.0.5/include/luajit-2.0",
    "W:/Projects/NervSeed/sources/nvCore/include"
  }

if false then

  -- We try to generate the simple PCH file:
  local pchfile = self:getPCHForBuffer([[
#include <core_lua.h>
#include <lua/LuaManager.h>
#include <NervApp.h>
#include <view/WindowManager.h>

using namespace nv;
  ]])

  local startTime = nv.SystemTime.getCurrentTime()
  logDEBUG("JIT: Loading Lua extensions...")
  self:usePCHFile(pchfile)

  self:loadBitcodeForFile(self.script_dir.."lua_base_extensions.cpp")

  -- self.jit:generateBitcodeFromFile(self.script_dir.."lua_base_extensions.cpp", self.bc_dir.."lua_base_extensions.bc")
  -- self.jit:loadModuleBitcode(self.bc_dir.."lua_base_extensions.bc")
  
  -- self.jit:loadModuleFromFile(self.script_dir.."lua_base_extensions.cpp")
  local endTime = nv.SystemTime.getCurrentTime()
  logDEBUG(string.format("Script compiled in %.3fms", (endTime - startTime)*1000.0))

  self.jit:call("loadLuaBaseExtensions")

end

  logDEBUG("Done initializing JITCompiler")
end

function Class:addDynamicLib(libfile)
  if self._dynLibs[libfile] then
    logDEBUG("Dynamic library ",libfile," is already loaded in JIT.")
    return
  end

  self._dynLibs[libfile] = true
  self.jit:addDynamicLib(libfile)
end

function Class:uninit()
  self.jit = nil
  -- collect the garbages:
  utils.gc()

  -- stop LLVM:
  nv.uninitLLVM()
end

function Class:getPCHForFile(filename, force)
  -- We want to generate a PCH for a given file, so we read the content of that file into a buffer:
  CHECK(nv.fileExists(filename), "File ", filename, " doesn't exist.")
  local buf = nv.readFile(filename)
  return self:getPCHForBuffer(buf, force)
end

function Class:getPCHForBuffer(buf, force)
  -- First we should compute the hash of the input buffer:
  local bufHash = self:getContextualBufferHash(buf)

  -- If we are forcing pch generation then we do not perform any check,
  -- Otherwise we check if we have a pch file already corresponding to that buf,
  -- and if not, we generate the pch with the given name.
  local pchfile = self.pch_dir..bufHash..".pch"
  if force or not nv.fileExists(pchfile) then
    logDEBUG("Generating PCH file ", pchfile, "...")
    local startTime = nv.SystemTime.getCurrentTime()
    self.jit:generatePCHFromBuffer(buf,pchfile)
    local endTime = nv.SystemTime.getCurrentTime()
    logDEBUG(string.format("PCH compiled in %.3fms", (endTime - startTime)*1000.0))
  end

  -- Finally we return the name of that pch:
  return pchfile
end

function Class:usePCHForFile(filename, force)
  local pch = self:getPCHForFile(filename)
  self:usePCHFile(pch)
end

function Class:usePCHForBuffer(buf, force)
  local pch = self:getPCHForBuffer(buf)
  self:usePCHFile(pch)
end

function Class:buildBitcodeForBuffer(buf, force)
  -- First we should compute the hash of the input buffer:
  local bufHash = self:getContextualBufferHash(buf)

  local bcfile = self.bc_dir..bufHash..".bc"
  if force or not nv.fileExists(bcfile) then
    logDEBUG("Generating bytecode file ", bcfile, "...")
    local startTime = nv.SystemTime.getCurrentTime()
    self.jit:generateBitcodeFromBuffer(buf, bcfile)
    local endTime = nv.SystemTime.getCurrentTime()
    logDEBUG(string.format("Bitcode compiled in %.3fms", (endTime - startTime)*1000.0))
  end

  return bcfile
end

function Class:buildBitcodeForFile(filename, force)
  CHECK(nv.fileExists(filename), "File ", filename, " doesn't exist.")
  local buf = nv.readFile(filename)
  return self:buildBitcodeForBuffer(buf, force)
end

function Class:loadBitcodeForFile(filename, force)
  local bcfile = self:buildBitcodeForFile(filename, force)
  -- Finally we load the generated bytecode:
  self.jit:loadModuleBitcode(bcfile)
end

function Class:loadBitcodeForBuffer(buf, force)
  local bcfile = self:buildBitcodeForBuffer(buf, force)
  -- Finally we load the generated bytecode:
  self.jit:loadModuleBitcode(bcfile)
end

function Class:findFile(file)
  if path.extension(file) == "" then
    file = file..".cpp"
  end

  if nv.fileExists(file) then
    return file
  end

  for k,v in ipairs(self.script_search_paths) do
    local fullpath = v..file
    if nv.fileExists(fullpath) then
      return fullpath
    end
  end

  THROW("Cannot find script file: ", file)
end

function Class:linkModule(files, force)
  -- For each file, we generate the bitcode:
  local bcfiles = {}
  for _,file in pairs(files) do 
    file = self:findFile(file)
    table.insert(bcfiles, self:buildBitcodeForFile(file, force))
  end

  local modHash = self:computeStringListHash(bcfiles)

  -- Now we should link the module and generate the corresponding hash:
  local modfile = self.bc_dir..modHash..".bc"

  if force or not nv.fileExists(modfile) then
    logDEBUG("Linking module file ", modfile, "...")
    local startTime = nv.SystemTime.getCurrentTime()
    self.jit:linkModule(modfile, bcfiles, true, true, true, true)
    local endTime = nv.SystemTime.getCurrentTime()
    logDEBUG(string.format("Linked module  in %.3fms", (endTime - startTime)*1000.0))
  end

  return modfile
end

function Class:loadModule(files, force)
  local modfile = self:linkModule(files, force)
  self.jit:loadModuleBitcode(modfile)
end

function Class:loadScript(sname, force)
  -- We should add the cpp extension:
  local file = self:findFile(sname)
  
  -- Then we load that file directly:
  self:loadBitcodeForFile(file, force)
end

function Class:executeScript(sname, funcName)

  -- The script name should not contain any forward slash:
  sname = sname:gsub("\\","/")

  if funcName == nil then
    -- If no function name is provided then we assume we should find a function
    -- in that script with the same name as the script itself.
    -- But we should discard any parent folder in that name:
    funcName = sname:gsub(".*/","")
  end

  -- Then we load that script:
  self:loadScript(sname)

  -- and we execute the function:
  self:execute(funcName)
end

function Class:execute(funcName)
  self.jit:call(funcName)
end

function Class:runFunction(funcName, code)
  self:loadBitcodeForBuffer(code)
  self:execute(funcName)
end

function Class:runLambda(code)
  -- We check if we have already registered a lambda with that code:
  local hash = self:getContextualBufferHash(code)
  local funcName = self.lambda_map[hash]

  if funcName == nil then
    -- We should generate a new lambda name:
    self.lambda_count = self.lambda_count+1
    funcName = string.format("lambda_func_%d", self.lambda_count)
    self.lambda_map[hash] = funcName
  end

  -- Now we encapsulate the code in the function:
  code = table.concat({"extern \"C\" void ", funcName, "() { ", code, " }"},"")
  self:runFunction(funcName, code)
end


function Class:clearBitcodeCache()
  -- get the list of all bytecode files in the compiled folder:
  local delFile = function(fname)
    -- logDEBUG("Found file: ",fname)
    local ext = path.extension(fname)
    if ext == ".bc" then
      logDEBUG("Deleting file: ",fname)
      file.delete(path.join(self.bc_dir, fname))
    end
  end

  utils.forEachFile(self.bc_dir, delFile)
end

-- We return an instance of that class:
return Class()
