#include <llvm_precomp.h>
#include <unordered_map>

#include <llvm/IR/Verifier.h>
#include <llvm/Transforms/IPO.h>
#include <llvm/IR/LegacyPassManager.h>
#include <llvm/Target/TargetMachine.h>
#include <llvm/Analysis/TargetLibraryInfo.h>
#include <llvm/Analysis/TargetTransformInfo.h>
#include <llvm/Transforms/IPO/PassManagerBuilder.h>
#include <clang/Frontend/FrontendActions.h>
#include <llvm/IRReader/IRReader.h>
#include <llvm/ExecutionEngine/Orc/ExecutionUtils.h>
#include "llvm/ExecutionEngine/Orc/ObjectLinkingLayer.h"
#include "llvm/ExecutionEngine/JITLink/EHFrameSupport.h"
#include "llvm/Linker/Linker.h"
#include "llvm/Bitcode/BitcodeWriter.h"
#include "llvm/Transforms/IPO/Internalize.h"

using namespace llvm::orc;
using namespace llvm;

#define NV_MAX_FUNCTION_NAME_LENGTH 256

#define NV_REGISTER_CXA_EXIT 0
#if defined(_WIN64)
// cf. https://reviews.llvm.org/D35103
#define NV_JIT_HANDLE_EXCEPTIONS 1
#endif

#ifdef NV_JIT_HANDLE_EXCEPTIONS
#include "llvm/Support/DynamicLibrary.h"

#define WIN32_LEAN_AND_MEAN
#define NOGDI
// #define NOMINMAX
#include <windows.h>
#endif

// cf. https://stackoverflow.com/questions/53738883/run-default-optimization-pipeline-using-modern-llvm/53739108

#define DEBUG_MSG(msg) std::cout << "[DEBUG]: "<<msg<< std::endl;
#define ERROR_MSG(msg) std::cout << "[ERROR]: "<<msg<< std::endl;

#define THROW_MSG(msg)                                                                          \
    {                                                                                           \
        ERROR_MSG(msg); \
        throw std::runtime_error("An exception just occured.");                                                     \
    }

namespace nv
{


#if NV_JIT_HANDLE_EXCEPTIONS
class SingleSectionMemoryManager : public llvm::SectionMemoryManager {
  struct Block {
    uint8_t *Addr = nullptr, *End = nullptr;
    void Reset(uint8_t *Ptr, uintptr_t Size);
    uint8_t *Next(uintptr_t Size, unsigned Alignment);
  };
  Block Code, ROData, RWData;

public:
  uint8_t *allocateCodeSection(uintptr_t Size, unsigned Align, unsigned ID,
                               llvm::StringRef Name) final;

  uint8_t *allocateDataSection(uintptr_t Size, unsigned Align, unsigned ID,
                               llvm::StringRef Name, bool RO) final;

  void reserveAllocationSpace(uintptr_t CodeSize, uint32_t CodeAlign,
                              uintptr_t ROSize, uint32_t ROAlign,
                              uintptr_t RWSize, uint32_t RWAlign) final;

  bool needsToReserveAllocationSpace() override { return true; }

#ifdef NV_JIT_HANDLE_EXCEPTIONS
  using llvm::SectionMemoryManager::EHFrameInfos;

  SingleSectionMemoryManager();

  void deregisterEHFrames() override;

  bool finalizeMemory(std::string *ErrMsg) override;

private:
  uintptr_t ImageBase = 0;
#endif
};

void SingleSectionMemoryManager::Block::Reset(uint8_t *Ptr, uintptr_t Size) {
  assert(Ptr != nullptr && "Bad allocation");
  Addr = Ptr;
  End = Ptr ? Ptr + Size : nullptr;
}

uint8_t *SingleSectionMemoryManager::Block::Next(uintptr_t Size,
                                                 unsigned Alignment) {
  uintptr_t Out = (uintptr_t)Addr;

  // Align the out pointer properly
  if (!Alignment)
    Alignment = 16;
  Out = (Out + Alignment - 1) & ~(uintptr_t)(Alignment - 1);

  // RuntimeDyld should have called reserveAllocationSpace with an amount that
  // will fit all required alignemnts...but assert on this to make sure.
  assert((Out + Size) <= (uintptr_t)End && "Out of bounds");

  // Set the next Addr to deliver at the end of this one.
  Addr = (uint8_t *)(Out + Size);
  return (uint8_t *)Out;
}

uint8_t *SingleSectionMemoryManager::allocateCodeSection(uintptr_t Size,
                                                         unsigned Align,
                                                         unsigned ID,
                                                         StringRef Name) {
  return Code.Next(Size, Align);
}

uint8_t *SingleSectionMemoryManager::allocateDataSection(
    uintptr_t Size, unsigned Align, unsigned ID, StringRef Name, bool RO) {
  return RO ? ROData.Next(Size, Align) : RWData.Next(Size, Align);
}

void SingleSectionMemoryManager::reserveAllocationSpace(
    uintptr_t CodeSize, uint32_t CodeAlign, uintptr_t ROSize, uint32_t ROAlign,
    uintptr_t RWSize, uint32_t RWAlign) {
  // FIXME: Ideally this should be one contiguous block, with Code, ROData,
  // and RWData pointing to sub-blocks within, but setting the correct
  // permissions for that wouldn't work unless we over-allocated to have each
  // Block.Base aligned on a page boundary.
  const unsigned SecID = 0;
  Code.Reset(SectionMemoryManager::allocateCodeSection(CodeSize, CodeAlign,
                                                       SecID, "code"),
             CodeSize);

  ROData.Reset(SectionMemoryManager::allocateDataSection(ROSize, ROAlign, SecID,
                                                         "rodata", true/*RO*/),
               ROSize);

  RWData.Reset(SectionMemoryManager::allocateDataSection(RWSize, RWAlign, SecID,
                                                         "rwdata", false/*RO*/),
               RWSize);

#ifdef NV_JIT_HANDLE_EXCEPTIONS
  ImageBase =
      (uintptr_t)std::min(std::min(Code.Addr, ROData.Addr), RWData.Addr);
#endif
}

#ifdef NV_JIT_HANDLE_EXCEPTIONS

// Map an "ImageBase" to a range of adresses that can throw.
//
class SEHFrameHandler {
  typedef SingleSectionMemoryManager::EHFrameInfos EHFrameInfos;
  typedef std::vector<std::pair<DWORD, DWORD>> ImageRanges;
  typedef std::map<uintptr_t, ImageRanges> ImageBaseMap;
  ImageBaseMap m_Map;

  static void MergeRanges(ImageRanges &Ranges);
  uintptr_t FindEHFrame(uintptr_t Caller);

public:
  static __declspec(noreturn) void __stdcall RaiseSEHException(void *, void *);
  void RegisterEHFrames(uintptr_t ImageBase, const EHFrameInfos &Frames,
                        bool Block = true);
  void DeRegisterEHFrames(uintptr_t ImageBase, const EHFrameInfos &Frames);
};

// Merge overlaping ranges for faster searching with throwing PC
void SEHFrameHandler::MergeRanges(ImageRanges &Ranges) {
  std::sort(Ranges.begin(), Ranges.end());

  ImageRanges Merged;
  ImageRanges::iterator It = Ranges.begin();
  auto Current = *(It)++;
  while (It != Ranges.end()) {
    if (Current.second + 1 < It->first) {
      Merged.push_back(Current);
      Current = *(It);
    } else
      Current.second = std::max(Current.second, It->second);
    ++It;
  }
  Merged.emplace_back(Current);
  Ranges.swap(Merged);
}

// Find the "ImageBase" for Caller/PC who is throwing an exception
uintptr_t SEHFrameHandler::FindEHFrame(uintptr_t Caller) {
  for (auto &&Itr : m_Map) {
    const uintptr_t ImgBase = Itr.first;
    for (auto &&Rng : Itr.second) {
      if (Caller >= (ImgBase + Rng.first) && Caller <= (ImgBase + Rng.second))
        return ImgBase;
    }
  }
  return 0;
}

// Register a range of adresses for a single section that
void SEHFrameHandler::RegisterEHFrames(uintptr_t ImageBase,
                                       const EHFrameInfos &Frames, bool Block) {
  if (Frames.empty())
    return;
  assert(m_Map.find(ImageBase) == m_Map.end());

  ImageBaseMap::mapped_type &Ranges = m_Map[ImageBase];
  ImageRanges::value_type *BlockRange = nullptr;
  if (Block) {
    // Merge all unwind adresses into a single contiguous block for faster
    // searching later.
    Ranges.emplace_back(std::numeric_limits<DWORD>::max(),
                        std::numeric_limits<DWORD>::min());
    BlockRange = &Ranges.back();
  }

  for (auto &&Frame : Frames) {
    assert(m_Map.find(DWORD64(Frame.Addr)) == m_Map.end() &&
           "Runtime function should not be a key!");

    PRUNTIME_FUNCTION RFunc = reinterpret_cast<PRUNTIME_FUNCTION>(Frame.Addr);
    const size_t N = Frame.Size / sizeof(RUNTIME_FUNCTION);
    if (BlockRange) {
      for (PRUNTIME_FUNCTION It = RFunc, End = RFunc + N; It < End; ++It) {
        BlockRange->first = std::min(BlockRange->first, It->BeginAddress);
        BlockRange->second = std::max(BlockRange->second, It->EndAddress);
      }
    } else {
      for (PRUNTIME_FUNCTION It = RFunc, End = RFunc + N; It < End; ++It)
        Ranges.emplace_back(It->BeginAddress, It->EndAddress);
    }

    ::RtlAddFunctionTable(RFunc, N, ImageBase);
  }

  if (!Block)
    MergeRanges(Ranges); // Initial sort and merge
}

void SEHFrameHandler::DeRegisterEHFrames(uintptr_t ImageBase,
                                         const EHFrameInfos &Frames) {
  if (Frames.empty())
    return;

  auto Itr = m_Map.find(ImageBase);
  if (Itr != m_Map.end()) {
    // Remove the ImageBase from lookup
    m_Map.erase(Itr);

    // Unregister all the PRUNTIME_FUNCTIONs
    for (auto &&Frame : Frames)
      ::RtlDeleteFunctionTable(reinterpret_cast<PRUNTIME_FUNCTION>(Frame.Addr));
  }
}

// FIXME: Rather than this static and overriding _CxxThrowException via
// DynamicLibrary::AddSymbol, a better route would be to transform the call
// to _CxxThrowException(Arg0, Arg1) -> RaiseSEHException(Arg0, Arg1, this)
// where 'this' is the SingleSectionMemoryManager instance.  This could probably
// be done with clang, and definitely possible by injecting an llvm-IR function
// into the module with the name '_CxxThrowException'
//
static SEHFrameHandler sFrameHandler;

void SingleSectionMemoryManager::deregisterEHFrames() {
  sFrameHandler.DeRegisterEHFrames(ImageBase, EHFrames);
  EHFrameInfos().swap(EHFrames);
}

bool SingleSectionMemoryManager::finalizeMemory(std::string *ErrMsg) {
  sFrameHandler.RegisterEHFrames(ImageBase, EHFrames);
  ImageBase = 0;
  return SectionMemoryManager::finalizeMemory(ErrMsg);
}

SingleSectionMemoryManager::SingleSectionMemoryManager() {
  // Override Windows _CxxThrowException to call into our local version that
  // can throw to and from the JIT.
//   that function should not be called here, instead, we register an absolute symbol in our JIT libs.
//   sys::DynamicLibrary::AddSymbol(
//       "_CxxThrowException",
//       (void *)(uintptr_t)&SEHFrameHandler::RaiseSEHException);
}

// Adapted from VisualStudio/VC/crt/src/vcruntime/throw.cpp
#ifdef _WIN64
#define _EH_RELATIVE_OFFSETS 1
#endif
// The NT Exception # that we use
#define EH_EXCEPTION_NUMBER ('msc' | 0xE0000000)
// The magic # identifying this version
#define EH_MAGIC_NUMBER1 0x19930520
#define EH_PURE_MAGIC_NUMBER1 0x01994000
// Number of parameters in exception record
#define EH_EXCEPTION_PARAMETERS 4

// A generic exception record
struct EHExceptionRecord {
  DWORD ExceptionCode;
  DWORD ExceptionFlags;               // Flags determined by NT
  _EXCEPTION_RECORD *ExceptionRecord; // Extra exception record (unused)
  void *ExceptionAddress;             // Address at which exception occurred
  DWORD NumberParameters; // No. of parameters = EH_EXCEPTION_PARAMETERS
  struct EHParameters {
    DWORD magicNumber;            // = EH_MAGIC_NUMBER1
    void *pExceptionObject;       // Pointer to the actual object thrown
    struct ThrowInfo *pThrowInfo; // Description of thrown object
#if _EH_RELATIVE_OFFSETS
    DWORD64 pThrowImageBase; // Image base of thrown object
#endif
  } params;
};

__declspec(noreturn) void __stdcall
SEHFrameHandler::RaiseSEHException(void *CxxExcept, void *Info) {
    // DEBUG_MSG("Entering SEHFrameHandler::RaiseSEHException()!");

  uintptr_t Caller;
  static_assert(sizeof(Caller) == sizeof(PVOID), "Size mismatch");

  USHORT Frames = CaptureStackBackTrace(1, 1, (PVOID *)&Caller, NULL);
  assert(Frames && "No frames captured");
  (void)Frames;

  const DWORD64 BaseAddr = sFrameHandler.FindEHFrame(Caller);
  if (BaseAddr == 0)
    _CxxThrowException(CxxExcept, (_ThrowInfo *)Info);

  // A generic exception record
  EHExceptionRecord Exception = {
      EH_EXCEPTION_NUMBER,      // Exception number
      EXCEPTION_NONCONTINUABLE, // Exception flags (we don't do resume)
      nullptr,                  // Additional record (none)
      nullptr,                  // Address of exception (OS fills in)
      EH_EXCEPTION_PARAMETERS,  // Number of parameters
      {EH_MAGIC_NUMBER1, CxxExcept, (struct ThrowInfo *)Info,
#if _EH_RELATIVE_OFFSETS
       BaseAddr
#endif
      }};

// const ThrowInfo* pTI = (const ThrowInfo*)Info;

#ifdef THROW_ISWINRT
  if (pTI && (THROW_ISWINRT((*pTI)))) {
    // The pointer to the ExceptionInfo structure is stored sizeof(void*)
    // infront of each WinRT Exception Info.
    ULONG_PTR *EPtr = *reinterpret_cast<ULONG_PTR **>(CxxExcept);
    EPtr--;

    WINRTEXCEPTIONINFO **ppWei = reinterpret_cast<WINRTEXCEPTIONINFO **>(EPtr);
    pTI = (*ppWei)->throwInfo;
    (*ppWei)->PrepareThrow(ppWei);
  }
#endif

  // If the throw info indicates this throw is from a pure region,
  // set the magic number to the Pure one, so only a pure-region
  // catch will see it.
  //
  // Also use the Pure magic number on Win64 if we were unable to
  // determine an image base, since that was the old way to determine
  // a pure throw, before the TI_IsPure bit was added to the FuncInfo
  // attributes field.
  if (Info != nullptr) {
#ifdef THROW_ISPURE
    if (THROW_ISPURE(*pTI))
      Exception.params.magicNumber = EH_PURE_MAGIC_NUMBER1;
#if _EH_RELATIVE_OFFSETS
    else
#endif // _EH_RELATIVE_OFFSETS
#endif // THROW_ISPURE

    // Not quite sure what this is about, but pThrowImageBase can never be 0
    // here, as that is used to mark when an "ImageBase" was not found.
#if 0 && _EH_RELATIVE_OFFSETS
    if (Exception.params.pThrowImageBase == 0)
      Exception.params.magicNumber = EH_PURE_MAGIC_NUMBER1;
#endif // _EH_RELATIVE_OFFSETS
  }

// Hand it off to the OS:
#if defined(_M_X64) && defined(_NTSUBSET_)
  RtlRaiseException((PEXCEPTION_RECORD)&Exception);
#else
  RaiseException(Exception.ExceptionCode, Exception.ExceptionFlags,
                 Exception.NumberParameters, (PULONG_PTR)&Exception.params);
#endif
}

#endif // NV_JIT_HANDLE_EXCEPTIONS

#endif


typedef std::unordered_map<std::string, std::string> FunctionNameMap;

struct NervJITImpl {
    std::unique_ptr<llvm::orc::LLJIT> lljit;
    clang::IntrusiveRefCntPtr<clang::DiagnosticOptions> diagnosticOptions;
    std::unique_ptr<clang::TextDiagnosticPrinter> textDiagnosticPrinter;
    clang::IntrusiveRefCntPtr<clang::DiagnosticIDs> diagIDs;
    clang::IntrusiveRefCntPtr<clang::DiagnosticsEngine> diagnosticsEngine;

    std::unique_ptr<llvm::orc::ThreadSafeContext> tsContext;
    std::unique_ptr<clang::CodeGenAction> action;
    std::unique_ptr<clang::GeneratePCHAction> gen_pch_action;
    std::unique_ptr<clang::EmitBCAction> emit_bc_action;
    std::unique_ptr<clang::CompilerInstance> compilerInstance;
    std::unique_ptr<llvm::orc::MangleAndInterner> mangler;

    std::unique_ptr<TargetMachine> targetMachine;
    void optimizeModule(llvm::Module *module);

    FunctionNameMap functionNames;

    char tmpFuncName[NV_MAX_FUNCTION_NAME_LENGTH];

    NervJITImpl();
    ~NervJITImpl();

    void setInputFile(const std::string& filename);
    void setInputBuffer(llvm::MemoryBuffer* buf);

    void generateBitcode(std::string outFile);
    void generatePCH(std::string outFile);

    void linkModule(const std::string& outFile, const std::vector<std::string>& inputList, bool onlyNeeded, bool internalize, bool optimize, bool preserveUseListOrder);

    void loadModule();
    void loadModule(std::unique_ptr<Module> module);

    void loadModuleBitcode(const std::string& bcfile);

    uint64_t lookup(const std::string& name);

    void setupCommandLine(const std::vector<std::string>& args);
};

void checkLLVMError(llvm::Error Err)
{
    if(Err) {
        THROW_MSG("LLVM error: "<<llvm::toString(std::move(Err)));
    }
}

template <typename T> T CHECK_LLVM(llvm::Expected<T> &&E) {
    checkLLVMError(E.takeError());
    return std::move(*E);
}

#if NV_REGISTER_CXA_EXIT
using DestructorPtr = void (*)(void *);
using CXXDestructorDataPair = std::pair<DestructorPtr, void *>;
using CXXDestructorDataPairList = std::vector<CXXDestructorDataPair>;

static CXXDestructorDataPairList DSOHandleOverride;

static int CXAAtExitOverride(DestructorPtr Destructor, void *Arg, void *DSOHandle)
{
    DEBUG_MSG("Registering destructor in my CXAAtExit override!");
    // auto& CXXDestructorDataPairs = *reinterpret_cast<CXXDestructorDataPairList*>(DSOHandle);
    // CXXDestructorDataPairs.push_back(std::make_pair(Destructor, Arg));
    return 0;
}
#endif

typedef void(*ExitFunc)();
std::vector<ExitFunc> exitFuncList;

static int AtExitOverride(ExitFunc func)
{
    DEBUG_MSG("Registering destructor in my AtExit override: "<<(const void*)func);
    exitFuncList.push_back(func);
    return 0;
}

static void runDestructors() {
//   auto& CXXDestructorDataPairs = DSOHandleOverride;
//   for (auto &P : CXXDestructorDataPairs)
//     P.first(P.second);
//   CXXDestructorDataPairs.clear();
    // We execute the functions in LIFO order:
    while(!exitFuncList.empty()) {
        DEBUG_MSG("Executing one at exit function.")
        exitFuncList.back()();
        exitFuncList.pop_back();
        DEBUG_MSG("Done executing one at exit function.")
    }
}

template <typename PtrTy> JITTargetAddress toTargetAddress(PtrTy *P) {
return static_cast<JITTargetAddress>(reinterpret_cast<uintptr_t>(P));
}

NervJITImpl::NervJITImpl()
{
    DEBUG_MSG("Creating target machine.");
    // auto& jtmb = builder.getJITTargetMachineBuilder();
    // CHECK(jtmb.hasValue(), "No JITTargetMachineBuilder available.");

    // Setting up target triple: Note that this doesn't need to happen for now on win64 ?
    //   const std::string TripleStr = llvm::sys::getProcessTriple();

    //   llvm::Triple T(TripleStr);

    //   // Use ELF on Windows-32 and MingW for now.
    // #ifndef CLANG_INTERPRETER_COFF_FORMAT
    //   if (T.isOSBinFormatCOFF())
    //     T.setObjectFormat(llvm::Triple::ELF);
    // #endif

    auto jtmb = CHECK_LLVM(JITTargetMachineBuilder::detectHost());

    // Resetting emulateTLS to false:
    auto& tgtOpts = jtmb.getOptions();
    tgtOpts.EmulatedTLS = false;
    tgtOpts.ExplicitEmulatedTLS = false;
    
    targetMachine = CHECK_LLVM(jtmb.createTargetMachine());
    DEBUG_MSG("Target machine using emulated TLS: "<<(targetMachine->useEmulatedTLS() ? "YES":"NO"));

    auto dl = CHECK_LLVM(jtmb.getDefaultDataLayoutForTarget());

    DEBUG_MSG("Default layout prefix for target: '" << dl.getGlobalPrefix() << "'");

    DEBUG_MSG("Creating LLJIT object.");

    LLJITBuilder llb;
    llb.setJITTargetMachineBuilder(std::move(jtmb)).setNumCompileThreads(2);
#if NV_JIT_HANDLE_EXCEPTIONS
    // We use our custom memory manager here:
    llb.setObjectLinkingLayerCreator([](ExecutionSession &ES, const Triple &triple) -> std::unique_ptr<ObjectLayer> {
        auto GetMemMgr = []() { return std::make_unique<SingleSectionMemoryManager>(); };
        auto ObjLinkingLayer = std::make_unique<RTDyldObjectLinkingLayer>(ES, std::move(GetMemMgr));
        
        // Note sure this is needed/appropriate ?
        if (triple.isOSBinFormatCOFF()) {
            ObjLinkingLayer->setOverrideObjectFlagsWithResponsibilityFlags(true);
            ObjLinkingLayer->setAutoClaimResponsibilityForObjectSymbols(true);
        }
        
        return std::unique_ptr<ObjectLayer>(std::move(ObjLinkingLayer));
    });
#endif

    lljit = CHECK_LLVM(llb.create());
    
    DEBUG_MSG("Done creating LLJIT object.")
    // cf. https://lists.llvm.org/pipermail/llvm-dev/2019-March/131057.html
    // cf. llvm/tools/lli/lli.cpp

    mangler = std::make_unique<llvm::orc::MangleAndInterner>(lljit->getExecutionSession(), lljit->getDataLayout());
    // runtimeOverrides = std::make_unique<llvm::orc::LocalCXXRuntimeOverrides>();

    // We assign the runtime overrides to out main dyn lib:
    // checkLLVMError(runtimeOverrides->enable(lljit->getMainJITDylib(), *mangler));

    {
        auto& JD = lljit->getMainJITDylib();
        SymbolMap RuntimeInterposes;
#if NV_REGISTER_CXA_EXIT
        // The __dso_handle and __cxa_atexit function should not be registerd with LLVM 11 implementation
        // since it is already provided internally.
        RuntimeInterposes[(*mangler)("__dso_handle")] =
            JITEvaluatedSymbol(toTargetAddress(&DSOHandleOverride),
                            JITSymbolFlags::Exported);
        RuntimeInterposes[(*mangler)("__cxa_atexit")] =
            JITEvaluatedSymbol(toTargetAddress(&CXAAtExitOverride),
                            JITSymbolFlags::Exported);
#endif
        // But we still need to manually take care of the atexit function itself:
        RuntimeInterposes[(*mangler)("atexit")] =
            JITEvaluatedSymbol(toTargetAddress(&AtExitOverride),
                            JITSymbolFlags::Exported);
#if NV_JIT_HANDLE_EXCEPTIONS
        RuntimeInterposes[(*mangler)("_CxxThrowException")] =
            JITEvaluatedSymbol(toTargetAddress(&SEHFrameHandler::RaiseSEHException),
                            JITSymbolFlags::Exported);
#endif

        checkLLVMError(JD.define(absoluteSymbols(std::move(RuntimeInterposes))));
    }

    diagnosticOptions = new clang::DiagnosticOptions;

    textDiagnosticPrinter = std::make_unique<clang::TextDiagnosticPrinter>(llvm::outs(), diagnosticOptions.get());

    // The diagnotic engine should not own the client below (or it could if we release our unique_ptr.)
    diagnosticsEngine = new clang::DiagnosticsEngine(diagIDs, diagnosticOptions, textDiagnosticPrinter.get(), false);

    compilerInstance = std::make_unique<clang::CompilerInstance>();
    auto& compilerInvocation = compilerInstance->getInvocation();

    // compilerInvocation.setLangDefaults(*langOptions, clang::InputKind(clang::Language::CXX), triple, );
    // auto* languageOptions = compilerInvocation.getLangOpts();
    // languageOptions->Exceptions = true;
    // languageOptions->CXXExceptions = true;

    // auto& preprocessorOptions = compilerInvocation.getPreprocessorOpts();
    auto& targetOptions = compilerInvocation.getTargetOpts();
    auto& frontEndOptions = compilerInvocation.getFrontendOpts();
    frontEndOptions.ShowStats = false;

    auto& headerSearchOptions = compilerInvocation.getHeaderSearchOpts();
    headerSearchOptions.Verbose = false;
    headerSearchOptions.UserEntries.clear();

    targetOptions.Triple = llvm::sys::getDefaultTargetTriple();
    compilerInstance->createDiagnostics(textDiagnosticPrinter.get(), false);

    std::unique_ptr<llvm::LLVMContext> context = std::make_unique<llvm::LLVMContext>();
    tsContext = std::make_unique<llvm::orc::ThreadSafeContext>(std::move(context));

    action = std::make_unique<clang::EmitLLVMOnlyAction>(tsContext->getContext());
    gen_pch_action = std::make_unique<clang::GeneratePCHAction>();
    emit_bc_action = std::make_unique<clang::EmitBCAction>();
}

NervJITImpl::~NervJITImpl()
{
    // We should uninitialize our main dyn library here:
    DEBUG_MSG("Uninitializing main JIT lib.");
    checkLLVMError(lljit->deinitialize(lljit->getMainJITDylib()));
    // checkLLVMError(lljit->runDestructors());
    // runtimeOverrides->runDestructors();
    runDestructors();
    DEBUG_MSG("Done uninitializing main JIT lib.");
}

NervJIT::NervJIT()
{
    DEBUG_MSG("Creating NervJIT object.");
    // impl = std::make_unique<NervJITImpl>();
    impl = new NervJITImpl();
    DEBUG_MSG("Done creating NervJIT object.");
}

NervJIT::~NervJIT()
{
    DEBUG_MSG("Deleting NervJIT object.");
//   impl.reset();
    delete impl;
    impl = nullptr;
    DEBUG_MSG("Deleted NervJIT object.");
}

void addOptPasses(
  llvm::legacy::PassManagerBase &passes,
  llvm::legacy::FunctionPassManager &fnPasses,
  llvm::TargetMachine *machine
) {
  llvm::PassManagerBuilder builder;
  builder.OptLevel = 3;
  builder.SizeLevel = 0;
  builder.Inliner = llvm::createFunctionInliningPass(3, 0, false);
  builder.LoopVectorize = true;
  builder.DisableUnrollLoops = false;
  builder.SLPVectorize = true;
  machine->adjustPassManager(builder);

  builder.populateFunctionPassManager(fnPasses);
  builder.populateModulePassManager(passes);
}

void addLinkPasses(llvm::legacy::PassManagerBase &passes) {
  llvm::PassManagerBuilder builder;
  builder.VerifyInput = true;
  builder.Inliner = llvm::createFunctionInliningPass(3, 0, false);
  builder.populateLTOPassManager(passes);
}

void NervJITImpl::optimizeModule(llvm::Module *module) {
    module->setTargetTriple(targetMachine->getTargetTriple().str());
    module->setDataLayout(targetMachine->createDataLayout());

    // DEBUG_MSG("Creating legacy pass manager.");
    llvm::legacy::PassManager passes;
    passes.add(new llvm::TargetLibraryInfoWrapperPass(targetMachine->getTargetTriple()));
    passes.add(llvm::createTargetTransformInfoWrapperPass(targetMachine->getTargetIRAnalysis()));

    // DEBUG_MSG("Creating legacy function pass manager.");
    llvm::legacy::FunctionPassManager fnPasses(module);
    fnPasses.add(llvm::createTargetTransformInfoWrapperPass(targetMachine->getTargetIRAnalysis()));

    addOptPasses(passes, fnPasses, targetMachine.get());
    addLinkPasses(passes);

    // DEBUG_MSG("Optimizing functions");
    fnPasses.doInitialization();
    for (llvm::Function &func : *module) {
        fnPasses.run(func);
    }
    fnPasses.doFinalization();

    passes.add(llvm::createVerifierPass());
    // DEBUG_MSG("Running PassManager.");
    passes.run(*module);
    // DEBUG_MSG("Optimization done.");
}

void NervJITImpl::generatePCH(std::string outFile)
{
    auto& compilerInvocation = compilerInstance->getInvocation();
    auto& frontEndOptions = compilerInvocation.getFrontendOpts();
    std::string prevFile = std::move(frontEndOptions.OutputFile);
    frontEndOptions.OutputFile = std::move(outFile);

    // keep a copy of the current program action:
    auto prevAction = frontEndOptions.ProgramAction;
    frontEndOptions.ProgramAction = clang::frontend::GeneratePCH;

    if (!compilerInstance->ExecuteAction(*gen_pch_action))
    {
        ERROR_MSG("Cannot execute gen_pch_action with compiler instance!");
    }

    // Restore the previous values:
    frontEndOptions.OutputFile = std::move(prevFile);
    frontEndOptions.ProgramAction = prevAction;
}

void NervJITImpl::generateBitcode(std::string outFile)
{
    auto& compilerInvocation = compilerInstance->getInvocation();
    auto& frontEndOptions = compilerInvocation.getFrontendOpts();
    std::string prevFile = std::move(frontEndOptions.OutputFile);
    frontEndOptions.OutputFile = std::move(outFile);

    // keep a copy of the current program action:
    auto prevAction = frontEndOptions.ProgramAction;
    frontEndOptions.ProgramAction = clang::frontend::EmitBC;

    if (!compilerInstance->ExecuteAction(*emit_bc_action))
    {
        ERROR_MSG("Cannot execute emit_bc_action with compiler instance!");
    }

    // Restore the previous values:
    frontEndOptions.OutputFile = std::move(prevFile);
    frontEndOptions.ProgramAction = prevAction;
}

void NervJITImpl::loadModule()
{
    if (!compilerInstance->ExecuteAction(*action))
    {
        ERROR_MSG("Cannot execute action with compiler instance!");
    }

    std::unique_ptr<llvm::Module> module = action->takeModule();
    if (!module)
    {
        ERROR_MSG("Cannot retrieve IR module.");
    }

    // We run the optimizations:
    DEBUG_MSG("Optimizing module...");
    optimizeModule(module.get());

    loadModule(std::move(module));
}


void NervJIT::loadModuleFromBuffer(const std::string& buffer)
{
    std::unique_ptr<MemoryBuffer> buf = llvm::MemoryBuffer::getMemBuffer(llvm::StringRef(buffer.data(), buffer.size()));
    impl->setInputBuffer(buf.get());
    impl->loadModule();
}

void NervJIT::loadModuleFromFile(const std::string& file)
{
    impl->setInputFile(file);
    impl->loadModule();
}

uint64_t NervJITImpl::lookup(const std::string& name)
{    
    // DEBUG_MSG("Searching symbol for "<<name);
    JITEvaluatedSymbol sym = CHECK_LLVM(lljit->lookupLinkerMangled(name.c_str()));
    // DEBUG_MSG("Returning address for "<<name);
    return sym.getAddress();
}

void NervJITImpl::setupCommandLine(const std::vector<std::string>& args)
{    
    std::vector<const char*> itemcstrs;
    for (unsigned idx = 0; idx < args.size(); idx++) {
        // DEBUG_MSG("Arg "<<idx<<": "<<args[idx].c_str());
        itemcstrs.push_back(args[idx].c_str());
    }

    auto& compilerInvocation = compilerInstance->getInvocation();
    if(!clang::CompilerInvocation::CreateFromArgs(compilerInvocation, llvm::ArrayRef<const char*>(itemcstrs.data(), itemcstrs.size()), *diagnosticsEngine.get()))
    {
        THROW_MSG("Cannot setup compiler invocation from args.");
    }

    // // Ensure the EmulateTLS is disabled:
    // auto& opts = compilerInvocation.getCodeGenOpts();
    // if(opts.EmulatedTLS) {
    //     DEBUG_MSG("Force disabling emulate TLS!")
    //     opts.EmulatedTLS = false;
    // }
    // if(opts.ExplicitEmulatedTLS) {
    //     DEBUG_MSG("Force disabling explicit emulate TLS!")
    //     opts.ExplicitEmulatedTLS = false;
    // }
}

void NervJITImpl::setInputFile(const std::string& filename)
{
    auto& compilerInvocation = compilerInstance->getInvocation();
    auto& frontEndOptions = compilerInvocation.getFrontendOpts();
    frontEndOptions.Inputs.clear();
    frontEndOptions.Inputs.push_back(clang::FrontendInputFile(llvm::StringRef(filename), clang::InputKind(clang::Language::CXX)));
}

void NervJITImpl::setInputBuffer(llvm::MemoryBuffer* buf)
{
    auto& compilerInvocation = compilerInstance->getInvocation();
    auto& frontEndOptions = compilerInvocation.getFrontendOpts();
    frontEndOptions.Inputs.clear();
    frontEndOptions.Inputs.push_back(clang::FrontendInputFile(buf, clang::InputKind(clang::Language::CXX)));
}

void NervJITImpl::loadModule(std::unique_ptr<Module> module)
{
    // cf. https://lists.llvm.org/pipermail/llvm-dev/2019-March/131057.html
    // llvm::orc::CtorDtorRunner R(lljit->getMainJITDylib());
    // R.add(llvm::orc::getConstructors(*module));

    DEBUG_MSG("Adding module from bytecode.");
    auto err = lljit->addIRModule(ThreadSafeModule(std::move(module), *tsContext));
    checkLLVMError(std::move(err));

    // Now we shoudl try to run the static initializers if any:
    // cf. https://groups.google.com/forum/#!msg/llvm-dev/DU5YYthVbrY/wXR1zZ7TAAAJ
    // DEBUG_MSG("Running module initializers.");
    DEBUG_MSG("Calling dyn lib initialize()");
    checkLLVMError(lljit->initialize(lljit->getMainJITDylib()));
    // checkLLVMError(R.run());
    DEBUG_MSG("Done calling dyn lib initialize()");
    // DEBUG_MSG("Done running module initializers.");
    // checkLLVMError(lljit->runConstructors());
}

void NervJITImpl::loadModuleBitcode(const std::string& bcfile)
{
    // Note we could also use getLazyIRFileModule here?
    llvm::SMDiagnostic Err;
    // llvm::LLVMContext Context;
    // std::unique_ptr<Module> Mod(llvm::parseIRFile(llvm::StringRef(bcfile), Err, Context));
    std::unique_ptr<Module> module(llvm::parseIRFile(llvm::StringRef(bcfile), Err, *tsContext->getContext()));
    if(!module) {
        THROW_MSG("Cannot load IR module from file "<<bcfile);
    }

    loadModule(std::move(module));
}

static void linkFiles(LLVMContext &Context, Linker &L,
                      const std::vector<std::string> &Files,
                      unsigned Flags, bool internalize) {

  // Filter out flags that don't apply to the first file we load.
  unsigned ApplicableFlags = Flags & Linker::Flags::OverrideFromSrc;
  // Similar to some flags, internalization doesn't apply to the first file.
  bool InternalizeLinkedSymbols = false;

  for (const auto &File : Files) {
    llvm::SMDiagnostic dErr;
    std::unique_ptr<Module> M(llvm::parseIRFile(llvm::StringRef(File), dErr, Context));

    if (!M.get()) {
        THROW_MSG("Cannot load module file: "<< File);
    }

    if(verifyModule(*M, &errs())) {
        THROW_MSG("Module file "<<File<<" is broken.");
    }

    bool Err = false;
    if (InternalizeLinkedSymbols) {
      Err = L.linkInModule(
          std::move(M), ApplicableFlags, [](Module &M, const StringSet<> &GVS) {
            internalizeModule(M, [&GVS](const GlobalValue &GV) {
              return !GV.hasName() || (GVS.count(GV.getName()) == 0);
            });
          });
    } else {
      Err = L.linkInModule(std::move(M), ApplicableFlags);
    }

    if (Err) {
        THROW_MSG("Error while linking in module.");
    }

    // Internalization applies to linking of subsequent files.
    InternalizeLinkedSymbols = internalize;

    // All linker flags apply to linking of subsequent files.
    ApplicableFlags = Flags;
  }
}

void NervJITImpl::linkModule(const std::string& outFile, const std::vector<std::string>& inputList, bool onlyNeeded, bool internalize, bool optimize, bool preserveUseListOrder)
{
    auto Composite = std::make_unique<Module>("llvm-link", *tsContext->getContext());
    llvm::Linker L(*Composite);

    unsigned Flags = Linker::Flags::None;
    if (onlyNeeded)
        Flags |= Linker::Flags::LinkOnlyNeeded;

    linkFiles(*tsContext->getContext(), L, inputList, Flags, internalize);

    if(verifyModule(*Composite, &errs())) {
        THROW_MSG("Composite module is broken.");
    }

    if(optimize) {
        optimizeModule(Composite.get());
    }

    // cf. https://stackoverflow.com/questions/13903686/writing-module-to-bc-bitcode-file
    std::error_code EC;
    llvm::raw_fd_ostream OS(outFile, EC, llvm::sys::fs::F_None);
    WriteBitcodeToFile(*Composite, OS, preserveUseListOrder);
    OS.flush();
}


uint64_t NervJIT::lookup(const std::string& name)
{
    return impl->lookup(name);
}

void NervJIT::clearHeaderSearchPaths()
{
    auto& compilerInvocation = impl->compilerInstance->getInvocation();
    auto& headerSearchOptions = compilerInvocation.getHeaderSearchOpts();
    headerSearchOptions.UserEntries.clear();
}

void NervJIT::addHeaderSearchPath(std::string path, HeaderType htype)
{
    auto& compilerInvocation = impl->compilerInstance->getInvocation();
    auto& headerSearchOptions = compilerInvocation.getHeaderSearchOpts();

    switch(htype)
    {
    case HEADER_SYSTEM:
        headerSearchOptions.AddPath(std::move(path), clang::frontend::System, false, false);
        break;
    case HEADER_QUOTED:
        headerSearchOptions.AddPath(std::move(path), clang::frontend::Quoted, false, false);
        break;
    case HEADER_ANGLED:
        headerSearchOptions.AddPath(std::move(path), clang::frontend::Angled, false, false);
        break;
    default:
        THROW_MSG("Unsupported header type: "<<htype);
    }
}

void NervJIT::clearMacroDefinitions()
{
    auto& compilerInvocation = impl->compilerInstance->getInvocation();
    auto& opts = compilerInvocation.getPreprocessorOpts();
    opts.Macros.clear();
}

void NervJIT::addMacroDefinition(std::string val)
{
    auto& compilerInvocation = impl->compilerInstance->getInvocation();
    auto& opts = compilerInvocation.getPreprocessorOpts();
    // We put our string directly into the vector instead of converting it to/from StringRef:
    opts.Macros.emplace_back(std::move(val), false);
}

void NervJIT::addDynamicLib(const char* filename)
{
    char prefix = impl->lljit->getDataLayout().getGlobalPrefix();
    // DEBUG_MSG("Loading dyn lib with prefix: '"<<prefix<<"'");
    impl->lljit->getMainJITDylib().addGenerator(cantFail(DynamicLibrarySearchGenerator::Load(filename, prefix)));
}

void NervJIT::addCurrentProcess()
{
    // cf. https://github.com/tensorflow/mlir/issues/24
    impl->lljit->getMainJITDylib().addGenerator(cantFail(DynamicLibrarySearchGenerator::GetForCurrentProcess(impl->lljit->getDataLayout().getGlobalPrefix())));
}

void NervJIT::setupCommandLine(const std::vector<std::string>& args)
{
    impl->setupCommandLine(args);
}

void NervJIT::generateBitcodeFromFile(const std::string& inputFile, std::string outFile)
{
    impl->setInputFile(inputFile);
    impl->generateBitcode(std::move(outFile));
}

void NervJIT::generateBitcodeFromBuffer(const std::string& buffer, std::string outFile)
{
    std::unique_ptr<MemoryBuffer> buf = llvm::MemoryBuffer::getMemBuffer(llvm::StringRef(buffer.data(), buffer.size()));
    impl->setInputBuffer(buf.get());
    impl->generateBitcode(std::move(outFile));
}


void NervJIT::generatePCHFromFile(const std::string& inputFile, std::string outFile)
{
    impl->setInputFile(inputFile);
    impl->generatePCH(std::move(outFile));
}

void NervJIT::generatePCHFromBuffer(const std::string& buffer, std::string outFile)
{
    std::unique_ptr<MemoryBuffer> buf = llvm::MemoryBuffer::getMemBuffer(llvm::StringRef(buffer.data(), buffer.size()));
    impl->setInputBuffer(buf.get());
    impl->generatePCH(std::move(outFile));
}

void NervJIT::usePCHFile(std::string pchfile)
{
    auto& compilerInvocation = impl->compilerInstance->getInvocation();
    auto& opts = compilerInvocation.getPreprocessorOpts();
    opts.ImplicitPCHInclude = std::move(pchfile);
}

void NervJIT::loadModuleBitcode(const std::string& bcfile)
{
    impl->loadModuleBitcode(bcfile);
}

void NervJIT::linkModule(const std::string& outFile, const std::vector<std::string>& inputList, bool onlyNeeded, bool internalize, bool optimize, bool preserveUseListOrder)
{
    impl->linkModule(outFile, inputList, onlyNeeded, internalize, optimize, preserveUseListOrder);
}

}
