#include <llvm_precomp.h>
#include "llvm/Support/InitLLVM.h"

#include <iostream>

#define NV_LLVM_VERBOSE 1


#ifdef ERROR_MSG
#undef ERROR_MSG
#endif
#ifdef DEBUG_MSG
#undef DEBUG_MSG
#endif

#define ERROR_MSG(msg) std::cout << "[ERROR]: "<<msg<< std::endl;
#define DEBUG_MSG(msg) std::cout << "[DEBUG]: "<<msg<< std::endl;

bool LLVMInitialized = false;

namespace nv {

void initLLVM()
{
    if (LLVMInitialized)
    {
        DEBUG_MSG("LLVM already initialized.");
        return;
    }

    // We have not initialized any pass managers for any device yet.
    // Run the global LLVM pass initialization functions.
    llvm::InitializeNativeTarget();
    llvm::InitializeNativeTargetAsmPrinter();
    llvm::InitializeNativeTargetAsmParser();

#if 0 
    auto& Registry = *llvm::PassRegistry::getPassRegistry();

    llvm::initializeCore(Registry);
    llvm::initializeScalarOpts(Registry);
    llvm::initializeVectorization(Registry);
    llvm::initializeIPO(Registry);
    llvm::initializeAnalysis(Registry);
    llvm::initializeTransformUtils(Registry);
    llvm::initializeInstCombine(Registry);
    llvm::initializeInstrumentation(Registry);
    llvm::initializeTarget(Registry);
#else
    // Initialize passes
    llvm::PassRegistry &Registry = *llvm::PassRegistry::getPassRegistry();
    llvm::initializeCore(Registry);
    llvm::initializeCoroutines(Registry);
    llvm::initializeScalarOpts(Registry);
    // llvm::initializeObjCARCOpts(Registry);
    llvm::initializeVectorization(Registry);
    llvm::initializeIPO(Registry);
    llvm::initializeAnalysis(Registry);
    llvm::initializeTransformUtils(Registry);
    llvm::initializeInstCombine(Registry);
    llvm::initializeAggressiveInstCombine(Registry);
    llvm::initializeInstrumentation(Registry);
    llvm::initializeTarget(Registry);
    // For codegen passes, only passes that do IR to IR transformation are
    // supported.
    llvm::initializeExpandMemCmpPassPass(Registry);
    llvm::initializeScalarizeMaskedMemIntrinPass(Registry);
    llvm::initializeCodeGenPreparePass(Registry);
    llvm::initializeAtomicExpandPass(Registry);
    llvm::initializeRewriteSymbolsLegacyPassPass(Registry);
    llvm::initializeWinEHPreparePass(Registry);
    llvm::initializeDwarfEHPreparePass(Registry);
    llvm::initializeSafeStackLegacyPassPass(Registry);
    llvm::initializeSjLjEHPreparePass(Registry);
    llvm::initializePreISelIntrinsicLoweringLegacyPassPass(Registry);
    llvm::initializeGlobalMergePass(Registry);
    llvm::initializeIndirectBrExpandPassPass(Registry);
    llvm::initializeInterleavedLoadCombinePass(Registry);
    llvm::initializeInterleavedAccessPass(Registry);
    llvm::initializeEntryExitInstrumenterPass(Registry);
    llvm::initializePostInlineEntryExitInstrumenterPass(Registry);
    llvm::initializeUnreachableBlockElimLegacyPassPass(Registry);
    llvm::initializeExpandReductionsPass(Registry);
    llvm::initializeWasmEHPreparePass(Registry);
    llvm::initializeWriteBitcodePassPass(Registry);
    llvm::initializeHardwareLoopsPass(Registry);
    llvm::initializeTypePromotionPass(Registry);
#endif

    LLVMInitialized = true;
    DEBUG_MSG("LLVM initialized.");
}

void uninitLLVM()
{
    if (!LLVMInitialized)
    {
        DEBUG_MSG("LLVM was not initialized.");
        return;
    }

    llvm::llvm_shutdown();
    LLVMInitialized = false;
    DEBUG_MSG("LLVM uninitialized.");
}

}

void runOptimizations(llvm::Module& module, bool debugPasses)
{
    DEBUG_MSG("Building pass builder.");
    llvm::PassBuilder passBuilder;
    llvm::LoopAnalysisManager loopAnalysisManager(debugPasses);
    llvm::FunctionAnalysisManager functionAnalysisManager(debugPasses);
    llvm::CGSCCAnalysisManager cGSCCAnalysisManager(debugPasses);
    llvm::ModuleAnalysisManager moduleAnalysisManager(debugPasses);
 
    passBuilder.registerModuleAnalyses(moduleAnalysisManager);
    passBuilder.registerCGSCCAnalyses(cGSCCAnalysisManager);
    passBuilder.registerFunctionAnalyses(functionAnalysisManager);
    passBuilder.registerLoopAnalyses(loopAnalysisManager);
    passBuilder.crossRegisterProxies(loopAnalysisManager, functionAnalysisManager, cGSCCAnalysisManager, moduleAnalysisManager);

    DEBUG_MSG("Running module pass manager.");
    llvm::ModulePassManager modulePassManager = passBuilder.buildPerModuleDefaultPipeline(llvm::PassBuilder::OptimizationLevel::O3);
    modulePassManager.run(module, moduleAnalysisManager);
    
    DEBUG_MSG("Done running module pass manager.");
}

void runClang(const nv::FileList& files)
{
    nv::initLLVM();
 
    clang::IntrusiveRefCntPtr<clang::DiagnosticOptions> diagnosticOptions = new clang::DiagnosticOptions;
    // clang::DiagnosticOptions diagnosticOptions;

    std::unique_ptr<clang::TextDiagnosticPrinter> textDiagnosticPrinter = std::make_unique<clang::TextDiagnosticPrinter>(llvm::outs(), diagnosticOptions.get());
    // std:: unique_ptr <clang::DiagnosticIDs> diagIDs;
    clang::IntrusiveRefCntPtr<clang::DiagnosticIDs> diagIDs;

    clang::IntrusiveRefCntPtr<clang::DiagnosticsEngine> diagnosticsEngine = new clang::DiagnosticsEngine(diagIDs, diagnosticOptions, textDiagnosticPrinter.get(), false);
 
    clang::CompilerInstance compilerInstance;
    auto& compilerInvocation = compilerInstance.getInvocation();

    std::stringstream ss;
    ss << "-triple=" << llvm::sys::getDefaultTargetTriple();
 
    std::istream_iterator<std::string> begin(ss);
    std::istream_iterator<std::string> end;
    std::istream_iterator<std::string> i = begin;
    std::vector<const char*> itemcstrs;
    std::vector<std::string> itemstrs;
    while (i != end) {
      itemstrs.push_back(*i);
      ++i;
    }

    for (unsigned idx = 0; idx < itemstrs.size(); idx++) {
      // note: if itemstrs is modified after this, itemcstrs will be full
      // of invalid pointers! Could make copies, but would have to clean up then...
      itemcstrs.push_back(itemstrs[idx].c_str());
    }
 
    // clang::CompilerInvocation::CreateFromArgs(compilerInvocation, itemcstrs.data(), itemcstrs.data() + itemcstrs.size(), *diagnosticsEngine.release());
    clang::CompilerInvocation::CreateFromArgs(compilerInvocation, llvm::ArrayRef<const char*>(itemcstrs.data(), itemcstrs.size()), *diagnosticsEngine.get());

    // auto* languageOptions = compilerInvocation.getLangOpts();
    // auto& preprocessorOptions = compilerInvocation.getPreprocessorOpts();
    auto& targetOptions = compilerInvocation.getTargetOpts();
    auto& frontEndOptions = compilerInvocation.getFrontendOpts();
#ifdef NV_LLVM_VERBOSE
    frontEndOptions.ShowStats = true;
#endif
    auto& headerSearchOptions = compilerInvocation.getHeaderSearchOpts();
#ifdef NV_LLVM_VERBOSE
    headerSearchOptions.Verbose = true;
#endif
    auto& codeGenOptions = compilerInvocation.getCodeGenOpts();

    // llvm::StringRef filename = "W:/Projects/NervSeed/temp/test1.cxx";
    // llvm::StringRef filename = file.c_str();

    frontEndOptions.Inputs.clear();
    // frontEndOptions.Inputs.push_back(clang::FrontendInputFile(filename, clang::InputKind(clang::Language::CXX)));

    for(auto& fname: files) {
        DEBUG_MSG("Adding input file: "<<fname);
        frontEndOptions.Inputs.push_back(clang::FrontendInputFile(llvm::StringRef(fname.c_str()), clang::InputKind(clang::Language::CXX)));
    }

    targetOptions.Triple = llvm::sys::getDefaultTargetTriple();
    compilerInstance.createDiagnostics(textDiagnosticPrinter.get(), false);
 

    llvm::LLVMContext context;
    std::unique_ptr<clang::CodeGenAction> action = std::make_unique<clang::EmitLLVMOnlyAction>(&context);
 
    if (!compilerInstance.ExecuteAction(*action))
    {
        ERROR_MSG("Cannot execute action with compiler instance.");
    }
 

    std::unique_ptr<llvm::Module> module = action->takeModule();
    if (!module)
    {
        ERROR_MSG("Cannot retrieve IR module.");
    }

    DEBUG_MSG("Before optimization module function list: ");
    int k=0;
    for(auto& f: *module)
    {
        DEBUG_MSG("func"<<k++<<": '"<<f.getName().str()<<"'");
    }

    bool debugPasses = false;
    codeGenOptions.DebugPassManager = debugPasses;

    runOptimizations(*module, debugPasses);
    
    // DEBUG_MSG("Building pass builder.");
    // llvm::PassBuilder passBuilder;
    // llvm::LoopAnalysisManager loopAnalysisManager(codeGenOptions.DebugPassManager);
    // llvm::FunctionAnalysisManager functionAnalysisManager(codeGenOptions.DebugPassManager);
    // llvm::CGSCCAnalysisManager cGSCCAnalysisManager(codeGenOptions.DebugPassManager);
    // llvm::ModuleAnalysisManager moduleAnalysisManager(codeGenOptions.DebugPassManager);
 
    // passBuilder.registerModuleAnalyses(moduleAnalysisManager);
    // passBuilder.registerCGSCCAnalyses(cGSCCAnalysisManager);
    // passBuilder.registerFunctionAnalyses(functionAnalysisManager);
    // passBuilder.registerLoopAnalyses(loopAnalysisManager);
    // passBuilder.crossRegisterProxies(loopAnalysisManager, functionAnalysisManager, cGSCCAnalysisManager, moduleAnalysisManager);

    // DEBUG_MSG("Running module pass manager.");
    // llvm::ModulePassManager modulePassManager = passBuilder.buildPerModuleDefaultPipeline(llvm::PassBuilder::OptimizationLevel::O3);
    // modulePassManager.run(*module, moduleAnalysisManager);

    DEBUG_MSG("Building engine builder.");
    llvm::EngineBuilder builder(std::move(module));
    builder.setMCJITMemoryManager(std::make_unique<llvm::SectionMemoryManager>());
    builder.setOptLevel(llvm::CodeGenOpt::Level::Aggressive);
    
    std::string createErrorMsg;
    // builder.setEngineKind(llvm::EngineKind::Interpreter);
    builder.setEngineKind(llvm::EngineKind::JIT);
    builder.setVerifyModules(true);
    builder.setErrorStr(&createErrorMsg);

    std::string triple = llvm::sys::getDefaultTargetTriple();
    DEBUG_MSG("Using target triple: "<<triple);
    auto executionEngine = builder.create();
 
    if (!executionEngine)
    {
        ERROR_MSG("Cannot create execution engine.'"<<createErrorMsg<<"'");
    }
    
    DEBUG_MSG("Retrieving nv_add/nv_sub functions...");
    typedef int(*AddFunc)(int,int);
    typedef int(*SubFunc)(int,int);

    AddFunc add = reinterpret_cast<AddFunc>(executionEngine->getFunctionAddress("nv_add"));
    if(!add) {
        ERROR_MSG("Cannot retrieve Add function.");
    }
    else {
        int res = add(40,2);
        ERROR_MSG("The meaning of life is: "<<res<<"!");
    }

    SubFunc sub = reinterpret_cast<SubFunc>(executionEngine->getFunctionAddress("nv_sub"));
    if(!sub) {
        ERROR_MSG("Cannot retrieve Sub function.");
    }
    else {
        int res = sub(50,8);
        ERROR_MSG("The meaning of life is really: "<<res<<"!");
    }


    DEBUG_MSG("leaving runClang() function.");
    // return reinterpret_cast<Function>(executionEngine->getFunctionAddress(function));
}